#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

const MAX_FILE_SIZE = 200000;


char hextable[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};


char* byteToHex(unsigned char inByte, char* buf)
{
	*buf = hextable[(inByte >> 4) & 15];
	*(buf+1) = hextable[inByte & 15];
	*(buf+2) = 0;
	return buf;
}

unsigned char hexToByte(char MSB, char LSB)
{
	unsigned char retbyte;
	if (MSB >= 'A' && MSB <= 'F')
		retbyte = (((unsigned char)MSB) - 'A' + 10) << 4;
	if (MSB >= '0' && MSB <= '9')
		retbyte = (((unsigned char)MSB) - '0') << 4;
	if (LSB >= 'A' && LSB <= 'F')
		retbyte |= (((unsigned char)LSB) - 'A' + 10);
	if (LSB >= '0' && LSB <= '9')
		retbyte |= (((unsigned char)LSB) - '0');
	return retbyte;
}



int getRawData(char* inData, int inDataSize, unsigned char* outData)
{
	int i, outDataSize=0;
	for (i = 0; i < inDataSize; i++)
	{

		// Are we looking at a hex byte?
		if (*inData == '$')
		{
			*outData = hexToByte(*(inData + 1), *(inData + 2));
			outData++;		// Skip the $XX characters
			i += 2;
			inData += 3;
			outDataSize++;	// Found another good byte, so use it
		}else{
			inData++;
		}
	}
	return outDataSize;
	
}





int rleCompress(unsigned char *data, int datasize, unsigned char *compressedData, unsigned char escapeByte)
{
	unsigned char	currentByte = *data, lastByte = *data;
	int				runLength=0,i=0,k,complen=0;
	unsigned char	writeLength;
	do{
		lastByte = currentByte;
		if (i != datasize)
			currentByte = *data;
		else
			currentByte = lastByte ^ 1;
		data++;
		if (currentByte == lastByte) 		// if it's a run
		{
			runLength++;					// increment the run counter
		}
		if (currentByte != lastByte || i == datasize )
		{
			if (runLength > 3)				// if there's a run worth using..
			{
				while (runLength > 0)
				{
					*compressedData = escapeByte;		// write the escape byte

					if (runLength > 255) {
						writeLength = 255;
					}else{
						writeLength = (unsigned char)runLength;
					}

					*(compressedData+1) = (unsigned char)(writeLength);	// write the length of the run
					
					*(compressedData+2) = lastByte;		// write the run's value
					compressedData += 3;				// and move to the next byte
					runLength -= 255;
					complen += 3;
				}
			}else{
				for (k=0;k<runLength;k++)
				{
					*(compressedData++) = lastByte;
					complen++;
				}
			}
			runLength = 1;
		}
		i++;
	}while (i <= datasize);
	return complen;
}


int main(int argc, char **argv)
{

	FILE	*openfile, *writefile;
	int		openfilesize, i;
	char	filecontents[MAX_FILE_SIZE];
	
	unsigned char
			rawdata[MAX_FILE_SIZE],
			compressdata[MAX_FILE_SIZE];

	if (argc < 4) {

		cout << "\nRLE Compressor\n\n   Usage:\n   rle infilename outfilename runflag\n   You MUST use file extensions.\n   runflag is a decimal value from 0-255\n";
		return 1;
	}


	if (!(openfile = fopen(argv[1], "rb"))) {
		cout << "Error opening infile " << argv[1] << "\nPlease make sure it exists.\n";
		return 1;
	}
	if (!(writefile = fopen(argv[2], "wb"))) {
		cout << "Error opening outfile " << argv[2] << "\nPlease make sure it exists.\n";
		return 1;
	}
	unsigned char escapebyte = (unsigned char) atoi(argv[3]);
	cout << "Compressing file \"" << argv[1] << "\" into file \"" << argv[2] << "\"...\n";




	char mybuf[5];

	openfilesize = fread(filecontents, 1, MAX_FILE_SIZE, openfile);
	int dataLen = getRawData(filecontents, openfilesize, (unsigned char*)rawdata);
	int compLen = rleCompress(rawdata, dataLen, compressdata, escapebyte);

	fprintf(writefile, ";RLE compressor by Joe Pemberton\r\n");
	fprintf(writefile, ";Compression: %i%% of the original size\r\n", compLen * 100 / dataLen);
	fprintf(writefile, ";Compressed size in bytes: %i\r\n", compLen);
	fprintf(writefile, ";Escape Character: $%s\r\n", byteToHex(escapebyte, mybuf));
	fprintf(writefile, "\r\n\t.dw %i\t\t\t;Uncompressed data size in bytes", dataLen);
	unsigned char *walkcompressdata = compressdata;
	for (i=0; i<compLen; i++)
	{
		if (!(i % 16))
			fprintf(writefile, "\r\n\t.db ");
		byteToHex(*(walkcompressdata++), mybuf);
		fprintf(writefile, "$%s", mybuf);
		if ((i % 16) != 15 && i < compLen - 1)
			fprintf(writefile, ",");

	}

	cout << "Compression successful!  Compressed data is in " << argv[2];
	cout << "\nRLE Compressor by Joe Pemberton\n  joe@joepnet.com\n  2003";
	fclose(writefile);
	fclose(openfile);
	return 0;
}


